import subprocess
import os
import sys
import argparse


def get_argparse() -> argparse.ArgumentParser:
    ap = argparse.ArgumentParser(
        description=__doc__
    )
    ap.add_argument(
        "-f",
        "--ffmpegPath",
        metavar="FFMPEG_PATH",
        help="Specific path to ffmpeg you want to use",
        default="ffmpeg",
    )
    ap.add_argument(
        "-r",
        "--secondsPerHash",
        metavar="NON_NEGATIVE_FLOAT",
        help="The frequence(per second) a hash is generated from the video. If it is 0, will generate every frame's hash",
        default="0",
        type=float,
    )
    ap.add_argument(
        "-d",
        "--outputHashFolder",
        metavar="OUTPUT_HASH_Folder_PATH",
        help="Output Hash Folder's Name",
        default="/ThreatExchange/vpdq/output-hashes",
        type=dir_path,
    )
    ap.add_argument(
        "-i",
        "--inputVideoFolder",
        metavar="INPUTPUT_VIDEO_FOLDER_PATH",
        help="Input Video Folder",
        default="/ThreatExchange/tmk/sample-videos",
        type=dir_path,
    )
    ap.add_argument(
        "-s",
        "--downsampleFrameDimension",
        metavar="Downsample_Frame_Dimension",
        help="Resolution to downsample the video to before hashing frames.. If it is 0, will use the original dimension of the video to hash",
        default="0",
        type=int,
    )
    ap.add_argument(
        "-t",
        "--matchDistanceTolerance",
        metavar="Matching_distanceTolerance",
        help="The hamming distance tolerance of between two frames. If the hamming distance is bigger than the tolerance, it will be considered as unmatched",
        default="10",
        type=int,
    )
    ap.add_argument(
        "-q",
        "--qualityTolerance",
        metavar="Matching_qualityTolerance",
        help="The quality tolerance of matching two frames. If either frames is below this quality level then they will not be compared",
        default="80",
        type=int,
    )
    ap.add_argument(
        "-v",
        "--verbose",
        help="If verbose, will print detailed information.",
        action="store_true",
    )
    return ap


def dir_path(string):
    if os.path.isdir(string):
        return string
    else:
        raise argparse.ArgumentTypeError(f"readable_dir: {string} is not a valid path")


def main():
    ap = get_argparse()
    args = ap.parse_args()
    inputVideoFolder = args.inputVideoFolder
    outputHashFolder = args.outputHashFolder
    ffmpegPath = args.ffmpegPath
    secondsPerHash = str(args.secondsPerHash)
    downsampleFrameDimension = str(args.downsampleFrameDimension)
    verbose = args.verbose
    # TODO: Add more general options for other video encodings.
    for file in os.listdir(inputVideoFolder):
        if file.endswith(".mp4"):
            if verbose:
                subprocess.call(
                    [
                        "./build/vpdq-hash-video",
                        "-v",
                        "-f",
                        ffmpegPath,
                        "-r",
                        secondsPerHash,
                        "-d",
                        outputHashFolder,
                        "-s",
                        downsampleFrameDimension,
                        "-i",
                        inputVideoFolder + "/" + file,
                    ]
                )
            else:
                subprocess.call(
                    [
                        "./build/vpdq-hash-video",
                        "-f",
                        ffmpegPath,
                        "-r",
                        secondsPerHash,
                        "-d",
                        outputHashFolder,
                        "-s",
                        downsampleFrameDimension,
                        "-i",
                        inputVideoFolder + "/" + file,
                    ]
                )

    cdir = os.getcwd()
    pdir = os.path.dirname(cdir)
    sample = pdir + ("/sample-hashes")
    output = outputHashFolder
    distanceTolerance = str(args.matchDistanceTolerance)
    qualityTolerance = str(args.qualityTolerance)
    for file in os.listdir(sample):
        if file.endswith(".txt"):
            print("\nMatching File " + file)
            sampleFile = f"{sample}/{file}"
            outputFile = f"{output}/{file}"
            if verbose:
                subprocess.call(
                    [
                        "./build/match-hashes-byline",
                        "-v",
                        sampleFile,
                        outputFile,
                        distanceTolerance,
                        qualityTolerance,
                    ]
                )
            else:
                subprocess.call(
                    [
                        "./build/match-hashes-byline",
                        sampleFile,
                        outputFile,
                        distanceTolerance,
                        qualityTolerance,
                    ]
                )


if __name__ == "__main__":
    main()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import vpdq


def read_file_to_hash(input_hash_filename: str) -> list[vpdq.VpdqFeature]:
    """Read hash file and return vpdq hash

    Args:
        input_hash_filename (str): Input hash file path

    Returns:
        list of VpdqFeature: vpdq hash from the hash file"""

    hash = []
    with open(input_hash_filename, "r") as file:
        lines = file.readlines()
    for line in lines:
        line = line.strip()
        content = line.split(",")
        pdq_hash = vpdq.str_to_hash(content[2])
        feature = vpdq.VpdqFeature(int(content[1]), int(content[0]), pdq_hash, float(content[3]))
        hash.append(feature)

    return hash


def output_hash_to_file(output_hash_filename, hashes):
    """Output vpdq hash to txt file

    Args:
        output_hash_filename (str): Output hash file path
        hash (list of VpdqFeature): Vpdq Hash"""

    with open(output_hash_filename, "w") as file:
        for cur in hashes:
            file.write(f"{cur.frame_number},{cur.quality},{hash_to_hex(cur.hash['w'])},{cur.timestamp:.3f}")
            file.write("\n")



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import vpdq
import pytest
import test_util
import os
from pathlib import Path
import sys

SAMPLE_HASH_FOLDER = Path("vpdq/sample-hashes")
SAMPLE_VIDEOS = Path("tmk/sample-videos")
DISTANCE_TOLERANCE = 10
QUALITY_TOLERANCE = 50
TEST_FILES = [
    "chair-19-sd-bar",
    "chair-20-sd-bar",
    "chair-22-sd-grey-bar",
    "chair-22-sd-sepia-bar",
    "chair-22-with-large-logo-bar",
    "chair-22-with-small-logo-bar",
    "chair-orig-22-fhd-no-bar",
    "chair-orig-22-hd-no-bar",
    "chair-orig-22-sd-bar",
    "doorknob-hd-no-bar",
    "pattern-hd-no-bar",
    "pattern-longer-no-bar",
    "pattern-sd-grey-bar",
    "pattern-sd-with-large-logo-bar",
    "pattern-sd-with-small-logo-bar",
]

project_dir = Path(__file__).parents[3]
hash_folder = project_dir / SAMPLE_HASH_FOLDER
video_folder = project_dir / SAMPLE_VIDEOS
test_hashes = {}
sample_hashes = {}


def test_vpdq_utils():
    sample = Path(f"{hash_folder}/{TEST_FILES[0]}.txt")
    assert sample.is_file()
    ret = test_util.read_file_to_hash(sample)
    with open(sample, "r") as file:
        lines = file.readlines()
    for line, feature in zip(lines, ret):
        line = line.strip()
        content = line.split(",")
        hex_hash = content[2]
        assert vpdq.str_to_hash(hex_hash) == feature.hash
        assert vpdq.hash_to_hex(feature.hash) == hex_hash


def test_error_checking():
    video_file = Path(f"{video_folder}/{TEST_FILES[0]}.mp4")
    with pytest.raises(ValueError, match="Seconds_per_hash must be non-negative"):
        vpdq.computeHash(input_video_filename=video_file, seconds_per_hash=-1)

    with pytest.raises(ValueError, match="Downsample_width must be non-negative"):
        vpdq.computeHash(input_video_filename=video_file, downsample_width=-1)

    with pytest.raises(ValueError, match="Downsample_height must be non-negative"):
        vpdq.computeHash(input_video_filename=video_file, downsample_height=-1)

    with pytest.raises(ValueError, match="Input_video_filename doesn't exist"):
        vpdq.computeHash(input_video_filename="nonexisting")


def test_compare_hashes():
    """This regression test is creating hashes from sample videos and compare them with the provided hashes line by line.
    Two VPDQ features are considered the same if each line of the hashes are within DISTANCE_TOLERANCE.
    For hashes that have a quality lower than QUALITY_TOLERANCE, the test will skip them for comoparing.
    """

    for file in TEST_FILES:
        hash_file = Path(f"{hash_folder}/{file}.txt")
        assert hash_file.is_file()
        ret = test_util.read_file_to_hash(hash_file)
        assert ret is not None
        sample_hashes[file] = ret

        video_file = Path(f"{video_folder}/{file}.mp4")
        assert video_file.is_file()
        ret = vpdq.computeHash(
            input_video_filename=video_file,
        )
        assert ret is not None
        test_hashes[file] = ret

        print("Comparing hash for video:", file)
        hash1 = test_hashes[file]
        hash2 = sample_hashes[file]
        assert len(hash1) == len(hash2)
        for h1, h2 in zip(hash1, hash2):
            if h1.quality >= QUALITY_TOLERANCE and h2.quality >= QUALITY_TOLERANCE:
                assert h1.hamming_distance(h2) < DISTANCE_TOLERANCE
                assert h1.frame_number == h2.frame_number



from setuptools import setup
from setuptools.extension import Extension
from setuptools.command.build_ext import build_ext                                     
import sys
import subprocess
from pathlib import Path

DIR = Path(__file__).parent
read_me = DIR / Path("vpdq/README.md")
long_description = read_me.read_text()
version = (DIR / "vpdq/version.txt").read_text(encoding="utf-8").strip()

class build_ext(build_ext):
    def run(self):
        command = ["make"]
        try:
            subprocess.check_call(command, cwd= DIR / "vpdq/cpp")
        except subprocess.CalledProcessError as e:
            print(e.output)
            print("fail to compile vpdq/pdq library")
            sys.exit(1)
        super().run()

EXTENSIONS = [
    Extension(
        "vpdq",
        ["vpdq/python/vpdq.pyx"],
        extra_objects=["vpdq/cpp/libvpdq.a"],
        include_dirs=["."],
        language="c++",
        extra_compile_args=["--std=c++11"],
    )
]

setup(
    name="vpdq",
    author="Facebook",
    description="Python bindings for Facebook VPDQ hash",
    author_email="threatexchange@fb.com",
    version=version,
    license_files="LICENSE.txt",
    license="BSD",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=[
        "cython",
        "opencv-python",
        "opencv-python-headless"
    ],
    include_package_data=True,
    cmdclass={'build_ext': build_ext},
    ext_modules=EXTENSIONS,
)



import shutil
from pathlib import Path
import os
import subprocess
import argparse


DIR = Path(__file__).resolve().parent
PARENTDIR = Path(__file__).resolve().parents[1]

SETUP = "setup.py"
MANIFEST = "MANIFEST.in"
DIST = "dist"


def get_argparse() -> argparse.ArgumentParser:
    ap = argparse.ArgumentParser()
    ap.add_argument(
        "-r",
        "--release",
        help="If release, will return the source distribution in the root dir.",
        action="store_true",
    )
    ap.add_argument(
        "-i",
        "--install",
        help="If install, will install vpdq locally.",
        action="store_true",
    )
    return ap


def main():
    ap = get_argparse()
    args = ap.parse_args()
    shutil.copy(DIR / SETUP, PARENTDIR / SETUP)
    shutil.copy(DIR / MANIFEST, PARENTDIR / MANIFEST)
    if args.release:
        print("build vpdq source distribution")
        run_command(["python3", "setup.py", "sdist", "bdist_wheel"], PARENTDIR)
        if (DIR / DIST).exists() and (DIR / DIST).is_dir():
            shutil.rmtree(DIR / DIST)
        shutil.move(PARENTDIR / DIST, DIR / DIST)
    if args.install:
        print("install vpdq source locally")
        run_command(["pip3", "install", "-e", "."], PARENTDIR)
    os.remove(PARENTDIR / SETUP)
    os.remove(PARENTDIR / MANIFEST)


def run_command(command, cwd="."):
    try:
        subprocess.check_call(command, cwd=cwd)
    except subprocess.CalledProcessError as e:
        print(e.output)


if __name__ == "__main__":
    main()



# distutils: language = c++
# cython: language_level=3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import cv2
import typing as t
import subprocess

from pathlib import Path
from dataclasses import dataclass
from libcpp cimport bool
from libcpp.vector cimport vector
from libcpp.string cimport string

try:
    # A call to check if ffmpeg is installed for vPDQ
    # FFMPEG is required to compute vPDQ hash
    subprocess.check_call(["ffmpeg", "-L"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
except FileNotFoundError as e:
    raise ImportError("FFMPEG is not installed.vPDQ requires FFMPEG. Visit https://ffmpeg.org/download.html to install. ")


cdef extern from "pdq/cpp/common/pdqhashtypes.h" namespace "facebook::pdq::hashing":
    cdef struct Hash256:
        unsigned short w[16];
    int hammingDistance(
        Hash256 hash1,
        Hash256 hash2
    )
    string hashToString(
        Hash256 hash
    )

cdef extern from "pdq/cpp/common/pdqhashtypes.h" namespace "facebook::pdq::hashing::Hash256":
    Hash256 fromStringOrDie(
        char* hex_str
    )

cdef extern from "vpdq/cpp/hashing/vpdqHashType.h" namespace "facebook::vpdq::hashing":
    cdef struct vpdqFeature:
        Hash256 pdqHash;
        int frameNumber;
        int quality;
        float timeStamp;


cdef extern from "vpdq/cpp/hashing/filehasher.h" namespace "facebook::vpdq::hashing":
    bool hashVideoFile(
        string input_video_filename,
        vector[vpdqFeature]& pdqHashes,
        string ffmpeg_path,
        bool verbose,
        double seconds_per_hash,
        int width,
        int height,
        double frames_per_sec,
        const char* argv0
    )

@dataclass
class VpdqFeature:
    quality: int
    frame_number: int
    hash: Hash256
    hex: str
    timestamp: double

    def __init__(
        self, quality: int, frame_number: int, hash: "Hash256", timestamp: double
    ):
        self.quality = quality
        self.frame_number = frame_number
        self.hash = hash
        self.hex = hash_to_hex(hash)
        self.timestamp = timestamp

    def hamming_distance(self, that: "vpdq_feature"):
        return hammingDistance(self.hash, that.hash)


def hash_to_hex(hash_value: "Hash256") -> str:
    """Convect from pdq hash to hex str

    Args:
        hash_value

    Returns:
        str: hex str of hash
    """
    return str(hashToString(hash_value), "utf-8")


def str_to_hash(str_hash: str):
    return fromStringOrDie(str(str_hash).encode("utf-8"))


def hamming_distance(hash1: "Hash256", hash2: "Hash256") -> int:
    """
    Return the hamming distance between two pdq hashes

    Args:
        hash1
        hash2

    Returns:
        int: The hamming distance
    """
    return hammingDistance(hash1, hash2)


def computeHash(
    input_video_filename: t.Union[str, Path],
    ffmpeg_path: str = "ffmpeg",
    seconds_per_hash: double = 0,
    verbose: bool = False,
    downsample_width: int = 0,
    downsample_height: int = 0,
):
    """Compute vpdq hash

    Args:
        input_video_filename: Input video file path
        ffmpeg_path: ffmpeg path
        verbose: If verbose, will print detailed information
        seconds_per_hash: The frequence(per second) a hash is generated from the video. If it is 0, will generate every frame's hash
        downsample_width: Width to downsample the video to before hashing frames.. If it is 0, will use the original width of the video to hash
        downsample_height: Height to downsample the video to before hashing frames.. If it is 0, will use the original height of the video to hash
    Returns:
        list of vpdq_feature: VPDQ hash from the video
    """
    str_path = str(input_video_filename)
    if not Path(str_path).is_file():
        raise ValueError("Input_video_filename doesn't exist")
    if seconds_per_hash < 0:
        raise ValueError("Seconds_per_hash must be non-negative")
    if downsample_width < 0:
        raise ValueError("Downsample_width must be non-negative")
    if downsample_height < 0:
        raise ValueError("Downsample_height must be non-negative")
    cdef vector[vpdqFeature] vpdq_hash;
    vid = cv2.VideoCapture(str_path)
    frames_per_sec = vid.get(cv2.CAP_PROP_FPS)
    if downsample_width == 0:
        downsample_width = vid.get(cv2.CAP_PROP_FRAME_WIDTH)

    if downsample_height == 0:
        downsample_height = vid.get(cv2.CAP_PROP_FRAME_HEIGHT)

    rt = hashVideoFile(
        str_path.encode("utf-8"),
        vpdq_hash,
        ffmpeg_path.encode("utf-8"),
        verbose,
        seconds_per_hash,
        downsample_width,
        downsample_height,
        frames_per_sec,
        "vpdqPY",
    )
    if not rt:
        raise Exception("Fail to create VPDQ hash")

    hashes = [
        VpdqFeature(hash.quality, hash.frameNumber, hash.pdqHash, hash.timeStamp)
        for hash in vpdq_hash
    ]
    return hashes
